from IPython.display import display, HTML
import pandas as pd
import numpy as np

from ...reference.Reference import Reference
from ...dbsignal.ExpSignalAnalysis import ExpSignalAnalysis


class EnergyExtractionAnalysis(object):

    @staticmethod
    def calculate_characteristic_time(df, t_start_decay, t_end_decay):
        index_start_decay = df.index.get_loc(t_start_decay, method='nearest')

        index_end_decay = df.index.get_loc(t_end_decay, method='nearest')
        df_decay = df.iloc[index_start_decay:index_end_decay, 0:1] - df.iloc[0].min()
        df_decay_squared = df_decay.pow(2)
        df_decay_tau = ExpSignalAnalysis.calculateCharacteristicTimeInSecondsWithIntegral(
            df_decay_squared, exponent=2)
        return df_decay_tau[0]

    @staticmethod
    def get_min_max_t_res_body_max(t_res_body_dfs):
        t_ref_body_max = [t_res_body_df.max().values[0] for t_res_body_df in t_res_body_dfs]

        return min(t_ref_body_max), max(t_ref_body_max)

    @staticmethod
    def rate_t_res_body(t_res_body_dfs):
        t_res_body_max_min, t_res_body_max_max = EnergyExtractionAnalysis.get_min_max_t_res_body_max(t_res_body_dfs)

        return "OK" if (t_res_body_max_max < 200) and (t_res_body_max_min > 100) else "NOT_OK"

    @staticmethod
    def analyse_energy_extraction(i_meas_df, u_dump_res_1_sync_df, u_dump_res_2_sync_df, t_delay_ee):
        index_i_meas_u_dump_res_1_max = i_meas_df.index.get_loc(u_dump_res_1_sync_df['U_DUMP_RES'].idxmax(),
                                                                method='nearest')
        i_meas_u_dump_res_1_max = i_meas_df.loc[i_meas_df.index[index_i_meas_u_dump_res_1_max], 'I_MEAS']

        index_i_meas_u_dump_res_2_max = i_meas_df.index.get_loc(u_dump_res_2_sync_df['U_DUMP_RES'].idxmax(),
                                                                method='nearest')
        i_meas_u_dump_res_2_max = i_meas_df.loc[i_meas_df.index[index_i_meas_u_dump_res_2_max], 'I_MEAS']

        tau_i_meas = EnergyExtractionAnalysis.calculate_characteristic_time(i_meas_df, t_start_decay=1,
                                                                            t_end_decay=120)
        tau_u_dump_res_1 = EnergyExtractionAnalysis.calculate_characteristic_time(u_dump_res_1_sync_df,
                                                                                  t_start_decay=1,
                                                                                  t_end_decay=120)
        tau_u_dump_res_2 = EnergyExtractionAnalysis.calculate_characteristic_time(u_dump_res_2_sync_df,
                                                                                  t_start_decay=1,
                                                                                  t_end_decay=120)


        values_act = [u_dump_res_1_sync_df['U_DUMP_RES'].max(),
                      u_dump_res_2_sync_df['U_DUMP_RES'].max(),
                      t_delay_ee[0] / 1e9,
                      t_delay_ee[1] / 1e9,
                      tau_i_meas,
                      tau_u_dump_res_1,
                      tau_u_dump_res_2]

        # ToDo: This could be resistance based ! - to a reference file
        values_min = [i_meas_u_dump_res_1_max * 0.075 * 0.9,
                      i_meas_u_dump_res_2_max * 0.075 * 0.9,
                      0.07,
                      0.052,
                      100,
                      100,
                      100]
        values_max = [i_meas_u_dump_res_1_max * 0.075 * 1.1,
                      i_meas_u_dump_res_2_max * 0.075 * 1.1,
                      0.13,
                      0.6,
                      110,
                      115,
                      115]
        columns = ['act', 'min', 'max']
        indices = ['U_DUMP_RES_1_max', 'U_DUMP_RES_2_max', 't_delay_trigger_1', 't_delay_trigger_2', 'tau_I_MEAS',
                   'tau_U_DUMP_RES_1', 'tau_U_DUMP_RES_2']

        ee_analysis_df = pd.DataFrame(index=indices, columns=columns, data=np.array([values_act, values_min, values_max]).T)
        ee_analysis_df['result'] = ee_analysis_df.apply(lambda row: (row['act'] >= row['min']) and (row['act'] <= row['max']),
                                                        axis=1)

        if all(ee_analysis_df['result'] == True):
            return True
        else:
            print('The following parameters did not satisfy the comparison criteria')
            display(HTML(ee_analysis_df[ee_analysis_df['result'] == False].to_html()))
            return False

    @staticmethod
    def analyse_energy_extraction_temperature(t_res_body_dfs):

        temp_ref_ee = Reference.getEnergyExtractionReferenceTemperature()

        # check if temperature is within the +/-25 range
        t_res_outside_range = []
        for index, t_res_body_df in enumerate(t_res_body_dfs):
            idx = temp_ref_ee.index.union(t_res_body_df.index)
            temp_ref_ee_analysis = temp_ref_ee.reindex(idx).interpolate('index')
            temp_ref_ee_analysis = temp_ref_ee_analysis.join(t_res_body_df).dropna()
            temp_ref_ee_analysis['Temp_min'] = temp_ref_ee_analysis['Temp'] - 25
            temp_ref_ee_analysis['Temp_max'] = temp_ref_ee_analysis['Temp'] + 25
            signal_name = t_res_body_df.columns[0]
            temp_ref_ee_analysis['Comparison'] = temp_ref_ee_analysis.apply(
                lambda row: (row[signal_name] >= row['Temp_min']) and (row[signal_name] <= row['Temp_max']), axis=1)
            if any(temp_ref_ee_analysis['Comparison'] == False):
                t_res_outside_range.append((index, signal_name))

        if len(t_res_outside_range) > 0:
            print('The following temperatures are outside of the reference profile by more than 25 K:')
            print(t_res_outside_range)

        # check if min_max_temperature
        t_res_max_outside_range = []
        t_ref_body_max = [t_res_body_df.max().values[0] for t_res_body_df in t_res_body_dfs]
        indices = [t_res_body_df.columns[0] for t_res_body_df in t_res_body_dfs]
        t_res_max_analysis_df = pd.DataFrame(index=indices, data=t_ref_body_max, columns=['T_RES_BODY_act'])
        t_res_max_analysis_df['T_RES_BODY_min'] = 100
        t_res_max_analysis_df['T_RES_BODY_max'] = 200

        t_res_max_analysis_df['Comparison'] = t_res_max_analysis_df.apply(lambda row: (
                    row['T_RES_BODY_act'] >= row['T_RES_BODY_min'] and (
                        row['T_RES_BODY_act'] <= row['T_RES_BODY_max'])), axis=1)

        if any(t_res_max_analysis_df['Comparison'] == False):
            print('The following maximum temperatures did not satisfy the comparison criteria:')
            display(HTML(t_res_max_analysis_df[t_res_max_analysis_df['result'] == False].to_html()))

        return (len(t_res_outside_range) == 0) and all(t_res_max_analysis_df['Comparison'] == True)

